import alice.tucson.api.*;
import alice.logictuple.*;
import java.util.*;


/**
 * Classe che implementa un automobilista capace di attraversare un incrocio con
 * semaforo e una rotonda.
 * La classe  un agente TuCSoN.
 */
public class AgenteAutomobile extends Agent {

/**
 * TupleCenter su cui si opera.
 */
	protected TupleCentreId tid;

/**
 * Posizione da cui proviene l'autista. Incapsula una stringa che pu assumere i
 * valori nord, sud, ovest, est.
 */
	protected Value origin;

/**
 * Posizione verso cui  diretto l'autista. Incapsula una stringa che pu 
 * assumere i valori nord, sud, ovest, est.
 */
	protected Value destination;

/**
 * Se good == true allora l'automobilista rispetter sempre le precedenza, 
 * altrimenti non rispetter il rosso qualora fosse fermo davanti al semaforo e
 * non dar la precedenza all'ingresso delle rotonde.
 */
	protected boolean good;

/**
 * Istante d'inizio della simulazione
 */
	protected long startTime;
	
/**
 * sim_time millisecondi del simulatore corrispondono ad 1 secondo reale.
 * All'aumentare di SIM_TIME diminuisce la velocit del simulatore.
 */	
	protected int sim_time;



/**
 * @param aid Stringa contenente l'identificativo dell'agente
 * @param t TupleCenter su cui dovr operare l'agente
 * @param o Stringa contenete la posizione iniziale dell'automobilista
 * @param d Stringa contenete la destinazione dell'automobilista
 * @param b E' vero se e solo se l'automobilista deve rispettare le precedenze
 * @param tm Tempo di simulazione (vedi sim_time)
 */
	public AgenteAutomobile (String aid, TupleCentreId t, String o, String d, boolean b, int tm) throws Exception {
		super(new AgentId(aid));
		tid = t;
		origin = new Value(o);
		destination = new Value(d);
		good = b;
		
		sim_time = tm;
	}
	
	
/**
 * Corpo dell'agente:
 * - ingresso in fila;
 * - attesa del proprio turno;
 * - attesa del semaforo verde (solo se si trova in un semaforo);
 * - attesa di uno spazio libero per passare e attraversamento dell'incrocio;
 * - uscita dall'incrocio.
 */
	protected void body(){
		Value tiket;
		try {
			tiket = enterQueue();
			waitForMyTurn(tiket);
			if (tid.getName().equals("semaforo")) waitForGreen();
			waitAndCross();
			exit();
		}
		catch (UnreachableNodeException une) {
			System.out.println("ERRORE: un agente non e' riuscito ad accedere a TuCSoN!");
		}
		catch (Exception ex) {
			ex.printStackTrace ();
		}
	}



/**
 * L'agente chiede il permesso di entrare nella coda dell'incrocio e riceve un
 * numero che corrisponde alla sua posizione nella fila.
 */
	protected Value enterQueue () throws Exception {
		Var x = new Var("X");
		in (tid, new LogicTuple("coda", origin, x));
		startTime = new Date().getTime();
		return new Value (Integer.parseInt (x.toString()));
	}
	
	
	
/**
 * L'agente aspetta di leggere sull'incrocio che tocca a lui, cio che tocca 
 * all'agente che possiede il 'biglietto' con il numero corrente 
 */
	protected void waitForMyTurn (Value tiket) throws Exception {
		rd (tid, new LogicTuple("avanti", origin, tiket));
	}
	
	
	
/**
 * Arrivato davanti al semaforo, l'agente 'buono' aspetta la luce verde, mentre 
 * quello 'cattivo' passa immediatamente.
 */
	protected void waitForGreen() throws Exception {
		if (good) { rd (tid, new LogicTuple("luce_verde", origin)); }
	}
	
	
	
/**
 * Procedura che consente all'agente di attraversare l'incrocio. E' divisa in due
 * parti esclusive a seconda che l'agente debba attraversare un incrocio con semaforo
 * o una rotonda.
 * Nel primo caso l'agente si limita a controllare che l'incrocio sia vuoto
 * (potrebbe essere occupato da un agente 'cattivo' che era passato col rosso) e
 * poi passa.
 * Nel caso della rotonda ci sono due comportamenti differenti a seconda che l'agente
 * sia 'buono' o .cattivo'. Se  'buono' aspetta che sia libero lo spazio davanti
 * a s e anche quello alla sua sinistra (da la precedenza a chi  gia dentro alla
 * rotonda), se  cattivo passa non appena il posto davanti a s  libero senza 
 * guardare a sinistra.
 * Di seguito il comportamento  uno solo e porta l'agente a percorrere tutta la
 * rotonda finch non si trova in corrispondenza della propria uscita.
 */
 	protected void waitAndCross() throws Exception {
		if (tid.getName().equals("semaforo")) {
			LogicTuple incrocio = new LogicTuple("spazio", new Value("incrocio"));
			in(tid, incrocio);
			out (tid, new LogicTuple("fatto", origin));
			doNothing (1 * sim_time);
			out(tid, incrocio);
		}


		if (tid.getName().equals("rotonda")) {
			Value curPos = origin;
			if (good) {
				in(tid, new LogicTuple("entrata", curPos));
				out(tid, new LogicTuple("spazio", getPrev(curPos), curPos));
			}
			else in(tid, new LogicTuple("spazio", curPos, getNext(curPos)));

			out (tid, new LogicTuple("fatto", origin));
			doNothing (1 * sim_time);
			curPos = getNext(curPos);
			
			while (!areSameDirection(curPos, destination)) {
				in(tid, new LogicTuple("spazio", curPos, getNext(curPos)));
				out(tid, new LogicTuple("spazio", getPrev(curPos), curPos));
				doNothing (1 * sim_time);
				curPos = getNext(curPos);
			}
			out(tid, new LogicTuple("spazio", getPrev(curPos), curPos));
		}
	}


/**
 * L'agente comunica all'ambiente di aver attraversato l'incrocio. Serve per fare
 * le statistiche.
 */
	protected void exit() throws Exception {
		out (tid, new LogicTuple("tempo_servizio", new Value(new Date().getTime() - startTime)));
	}



/**
 * Interrompe l'esecuzione per t millisecondi. In realt il programma guarda 
 * l'orologio finch non sono passati t millisecondi.
 */
	protected void doNothing (long t) {
		long start = new java.util.Date().getTime();
		long current;
		do {
			current = new java.util.Date().getTime();
		} while (start + t > current);
	}


	
/**
 * Dato un dato di tipo Value che incapsula una posizione (nord, sud, ovest, est)
 * restituisce un altro Value contenente la posizione successiva in senso antiorario.
 */
	protected Value getNext (Value v) {
		String s = v.toString();

		Value nord = new Value("nord");
		Value sud = new Value("sud");
		Value ovest = new Value("ovest");
		Value est = new Value("est");

		if (s.equals("nord")) return ovest;
		if (s.equals("est")) return nord;
		if (s.equals("sud")) return est;
		if (s.equals("ovest")) return sud;
		return null;
	}


	
/**
 * Dato un dato di tipo Value che incapsula una posizione (nord, sud, ovest, est)
 * restituisce un altro Value contenente la posizione precedente in senso antiorario.
 */
	protected Value getPrev (Value v) {
		String s = v.toString();

		Value nord = new Value("nord");
		Value sud = new Value("sud");
		Value ovest = new Value("ovest");
		Value est = new Value("est");

		if (s.equals("nord")) return est;
		if (s.equals("est")) return sud;
		if (s.equals("sud")) return ovest;
		if (s.equals("ovest")) return nord;
		return null;
	}
	
	
	
/**
 * Dato due dati di tipo Value che incapsulano una posizione (nord, sud, ovest, 
 * est), controlla se essi contengono la stessa posizione.
 */
	protected boolean areSameDirection (Value a, Value b) {
		String as = a.toString();
		String bs = b.toString();
		return as.equals(bs);
	}
}
